package type2;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import auxiliary.ListofListContains;
import co_reachable.CoReachability;
import reachable.Reachability;

public class Type2GrammarGenerator {

	Random random = new Random();

	private Map<String, List<List<String>>> genType2GrammarRuleSet(String[] vars, int alphaSize, int maxConclusios,
			int maxConclusioSize) {
		Map<String, List<List<String>>> rules = new HashMap<>();
		for (String v : vars) {
			List<List<String>> conclusios = genConclusios(alphaSize, vars, maxConclusios, maxConclusioSize);
			rules.put(v, conclusios);
		}
		return rules;
	}

	private List<List<String>> genConclusios(int alphaSize, String[] vars, int maxConclusios, int maxConclusioSize) {
		List<List<String>> conclusios = new LinkedList<>();
		int numOfConclusios = random.nextInt(maxConclusios) + 1;
		while (conclusios.size() < numOfConclusios) {
			List<String> conclusio = genConclusio(alphaSize, vars, maxConclusioSize);
			if (!ListofListContains.contains(conclusios, conclusio)) {
				conclusios.add(conclusio);
			}
		}
		return conclusios;
	}

	private List<String> genConclusio(int alphaSize, String[] vars, int maxConclusioSize) {
		return (Math.random() <= 0.1) ? new LinkedList<>() : genRandomString(alphaSize, vars, maxConclusioSize);

	}

	private List<String> genRandomString(int alphaSize, String[] vars, int maxConclusioSize) {
		List<String> s = new LinkedList<>();
		int conclusioSize = random.nextInt(maxConclusioSize) + 1;
		while (s.size() < conclusioSize) {
			int i = random.nextInt(vars.length + alphaSize);
			if (i < vars.length) {
				if (i == 0) {
					i = 83;
				} else if (i > 0 && i <= 18) {
					i += 64;
				} else {
					i += 65;
				}
			} else {
				i += 97 - vars.length;
			}
			char c = (char) i;
			s.add(c + "");
		}
		return s;
	}

	public List<Map<String, List<List<String>>>> genType2GrammarRuleSets(String[] vars, int alphaSize,
			int maxConclusios, int maxConclusioSize, int n) {
		List<Map<String, List<List<String>>>> grammarRuleSets = new LinkedList<>();
		while (grammarRuleSets.size() < n) {
			grammarRuleSets.add(genType2GrammarRuleSet(vars, alphaSize, maxConclusios, maxConclusioSize));
		}
		return grammarRuleSets;
	}

	public List<Map<String, List<List<String>>>> genType2GrammarRuleSets1(String[] vars, int alphaSize,
			int maxConclusios, int maxConclusioSize, String s, int n) {
		List<Map<String, List<List<String>>>> grammarRuleSets = new LinkedList<>();
		while (grammarRuleSets.size() < n) {
			Map<String, List<List<String>>> rules = genType2GrammarRuleSet(vars, alphaSize, maxConclusios,
					maxConclusioSize);
			if (!CoReachability.emptyLanguage(rules, s)) {
				grammarRuleSets.add(rules);
			}
		}
		return grammarRuleSets;
	}

	public List<Map<String, List<List<String>>>> genType2GrammarRuleSets2(String[] vars, int alphaSize,
			int maxConclusios, int maxConclusioSize, int n) {
		List<Map<String, List<List<String>>>> grammarRuleSets = new LinkedList<>();
		while (grammarRuleSets.size() < n) {
			Map<String, List<List<String>>> rules = genType2GrammarRuleSet(vars, alphaSize, maxConclusios,
					maxConclusioSize);
			if (CoReachability.coReachableSet(rules).containsAll(rules.keySet())) {
				grammarRuleSets.add(rules);
			}
		}
		return grammarRuleSets;
	}

	public List<Map<String, List<List<String>>>> genType2GrammarRuleSets3(String[] vars, int alphaSize,
			int maxConclusios, int maxConclusioSize, String s, int n) {
		List<Map<String, List<List<String>>>> grammarRuleSets = new LinkedList<>();
		Set<String> varSet = new HashSet<>();
		Set<String> alphaSet = new HashSet<>();
		for (String v : vars) {
			varSet.add(v);
		}
		for (int i = 0; i < alphaSize; i++) {
			alphaSet.add(((char) (i + 97)) + "");
		}
		while (grammarRuleSets.size() < n) {
			Map<String, List<List<String>>> rules = genType2GrammarRuleSet(vars, alphaSize, maxConclusios,
					maxConclusioSize);
			boolean allCoReachable = CoReachability.coReachableSet(rules).containsAll(rules.keySet());
			boolean allReachable = Reachability.reachableSet(rules, s).containsAll(rules.keySet())
					&& Reachability.reachableSet(rules, s).containsAll(alphaSet);
			if (allCoReachable && allReachable) {
				grammarRuleSets.add(rules);
			}
		}
		return grammarRuleSets;
	}

	public static void main(String[] args) {
		Map<String, List<List<String>>> ruleSet = new HashMap<>();
		List<List<String>> l1 = List.of(List.of("A", "S"));
		ruleSet.put("S", l1);
		List<List<String>> l2 = List.of(List.of("a"), List.of("b", "A"));
		ruleSet.put("A", l2);
		System.out.println(ruleSet);
		Set<String> crs = CoReachability.coReachableSet(ruleSet);
		System.out.println(crs);
	}

}
